import { self, pipe, exactly, length, either, string, number, run, both, even, iterableOf, str2num, fix, arrayOf, stringToDate, invalidDate, optional, not, atMost, object, empty, atLeast, Validator, createOneWayParser, stringOf, numberOf, createSimpleValidator, errorsPerKey, moreThan, when, boolean } from '../validators'
import { Message, enUS, describePaths } from '../message'

describe('validators', () => {
  test('string', async () => {
    expect(await run(string, 'test')).toEqual([
      [],
      'test',
      1
    ])
  })

  test('string or number / string', async () => {
    expect(await run(either(string, number), 'test')).toEqual([
      [],
      'test',
      1
    ])
  })

  test('string or number / number', async () => {
    expect(await run(either(string, number), 123)).toEqual([
      [],
      123,
      1
    ])
  })

  test('string or number / neither', async () => {
    expect(await run(either(string, number), {})).toEqual([
      ['Must be a string or a number']
      ,
      '',
      0
    ])
  })

  const w = both(number, even)

  test('conjunction', async () => {
    expect(await run(w, 123)).toEqual([
      ['Must be even'],
      123,
      0.75,
    ])
  })

  test('iterableOf / ok', async () => {
    expect(await run(iterableOf(str2num), ['1', '2', '3'])).toEqual([
      [],
      [1, 2, 3,],
      1,
    ])
  })

  test('iterableOf / fail', async () => {
    expect(await run(iterableOf(str2num), ['1', '', '3'])).toEqual([
      ['Must be an iterable where each element is a number'],
      [1, 0, 3,],
      2 / 3
    ])
  })

  test('arrayOf / fail 1', async () => {
    expect(await run(arrayOf(str2num), ['1', '', '3'])).toEqual([
      ['Must be an array where each element is a number'],
      [1, 0, 3,],
      2 / 3
    ])
  })

  test('arrayOf / fail 2', async () => {
    expect(await run(arrayOf(either(string, number)), null)).toEqual([
      ['Must be an array'],
      [],
      0,
    ])
  })

  test('arrayOf / fail', async () => {
    expect(await run(arrayOf(either(string, number)), [null])).toEqual([
      ['Must be an array where each element is a string or a number'],
      [''],
      0,
    ])
  })

  test('arrayOf / object 1', async () => {
    expect(await run(arrayOf(object({ a: number, b: string })), [0])).toEqual([
      ['Must be an array where each element is an object'],
      [{ a: 0, b: '' }],
      0
    ])
  })

  test('arrayOf / object 2', async () => {
    expect(await run(arrayOf(object({ a: number, b: string })), [{}])).toEqual([
      ['Must be an array where each element is an object where a must be a number'],
      [
        {
          a: 0,
          b: ''
        }
      ],
      0
    ])
  })

  test('arrayOf / object 2', async () => {
    expect(await run(arrayOf(either(number, object({ a: number, b: string }))), [{}])).toEqual([
      ['Must be an array where each element is a number or an object where a must be a number'],
      [0],
      0
    ])
  })

  type RecI = string | RecI[]
  type RecO = number | RecO[]
  const z = fix<RecI, RecO>(v => either(str2num, arrayOf(v)))

  test('fix / base case', async () => {
    expect(await run(z, '123')).toEqual([
      [],
      123,
      1
    ])
  })

  test('fix / one level of nesting', async () => {
    expect(await run(z, ['123'])).toEqual([
      [],
      [123],
      1
    ])
  })

  test('fix / multiple nest / valid', async () => {
    expect(await run(z, ['1', ['2', ['3'], '4', ['5', ['6']]]])).toEqual([
      [],
      [1, [2, [3], 4, [5, [6]]]],
      1
    ])
  })

  test('fix / multiple nest / invalid', async () => {
    const z = fix<RecI, RecO>(v => either(str2num, arrayOf(v)))

    expect(await run(z, ['1', ['2', ['3'], {} as any, ['5', ['6']]]])).toEqual([
      ['Must be a string or an array where each element is of this schema type'],
      [1, [2, [3], 0, [5, [6]]]],
      0.875
    ])
  })

  test('fix / multiple nest / validation failure', async () => {
    type Rec = number | RecO[]
    const z = fix<Rec>(v => either(even, arrayOf(v)))

    expect(await run(z, [1, [2, [3], {} as any, [5, [6]]]])).toEqual([
      ['Must be a number or an array where each element is of this schema type'],
      [1, [2, [3], 0, [5, [6]]]],
      0.53125
    ])
  })

  test('date / ok', async () => {
    expect(await run(stringToDate, '2018-11-18')).toEqual([
      [],
      new Date('2018-11-18'),
      1
    ])
  })

  test('date / invalid', async () => {
    expect(await run(stringToDate, '2018-20-18')).toEqual([
      ['Must be a date'],
      invalidDate,
      0
    ])
  })

  test('date / invalid 2', async () => {
    expect(await run(stringToDate, null as any)).toEqual([
      ['Must be a string'],
      invalidDate,
      0
    ])
  })

  test('optional / null', async () => {
    expect(await run(optional(stringToDate), null)).toEqual(
      [[], null, 1]
    )
  })

  test('optional / undefined', async () => {
    expect(await run(optional(stringToDate), undefined)).toEqual(
      [[], undefined, 1]
    )
  })

  test('optional / defined', async () => {
    expect(await run(optional(stringToDate), '2020-12-31')).toEqual(
      [[], new Date('2020-12-31'), 1]
    )
  })

  test('optional / invalid', async () => {
    expect(await run(optional(stringToDate), '2020-20-31')).toEqual([
      ['Must be a date'],
      invalidDate,
      0
    ])
  })
})

describe('not', () => {
  test('parser / valid', async () => {
    expect(await run(not(string), 'test')).toEqual([
      [],
      'test',
      1
    ])
  })

  test('parser / invalid', async () => {
    expect(await run(not(string), 3)).toEqual([
      ['Must be a string'],
      '',
      0
    ])
  })

  test('validator / invalid', async () => {
    expect(await run(not(even), 2)).toEqual([
      ['Must not be even'],
      2,
      0.5
    ])
  })

  test('validator / valid', async () => {
    expect(await run(not(even), 3)).toEqual([
      [],
      3,
      1,
    ])
  })

  test('not(disjunction) is conjunction', async () => {
    expect(await run(not(either(string, number)), 123)).toEqual([
      ['Must be a string'],
      123,
      0.5
    ])
  })

  test('conjunction / ok 1', async () => {
    expect(await run(not(both(even, atMost(10))), 123)).toEqual([
      [],
      123,
      1
    ])
  })

  test('conjunction / ok with negations', async () => {
    expect(await run(both(not(even), not(atMost(10))), 122)).toEqual([
      ['Must not be even'],
      122,
      0.75
    ])
  })

  test('conjunction / ok with top-level negation', async () => {
    expect(await run(not(both(even, atMost(10))), 122)).toEqual([
      [],
      122,
      1
    ])
  })

  test('conjunction / invalid', async () => {
    expect(await run(not(both(even, atMost(10))), 8)).toEqual([
      // TODO: Hard to understand?
      ['Must not be even or at most 10'],
      8,
      0.5
    ])
  })

  test('not iterableOf / ok', async () => {
    const v = not(iterableOf(even))
    expect(await run(v, [3, 1, 9] as any)).toEqual([
      [],
      [3, 1, 9],
      1
    ])
  })

  test('not iterableOf / failure', async () => {
    const v = not(iterableOf(even))
    expect(await run(v, ['a', 1] as any)).toEqual([
      ['Must be an iterable where each element is a number'],
      [0, 1],
      0.5
    ])
  })
})

describe('object', () => {
  test('ok', async () => {
    const v = object({
      a: even,
      b: string
    })
    expect(await run(v, { a: 10, b: 'test' } as any)).toEqual([
      [],
      {
        a: 10,
        b: 'test'
      },
      1
    ])
  })

  test('parse failure', async () => {
    const v = object({
      a: even,
      b: string
    })
    expect(await run(v, {} as any)).toEqual([
      [
        'a must be a number',
        'b must be a string'
      ],
      {
        a: 0,
        b: ''
      },
      0
    ])
  })

  test('one validation error', async () => {
    const v = object({
      a: even,
      b: string
    })
    expect(await run(v, { a: 11, b: 'test' } as any)).toEqual([
      ['a must be even'],
      {
        a: 11,
        b: 'test'
      },
      0.75
    ])
  })

  test('two validation errors', async () => {
    const v = object({
      a: even,
      b: not(empty)
    })
    expect(await run(v, { a: 11, b: '' } as any)).toEqual([
      [
        'a must be even',
        'b must not be empty',
      ],
      {
        a: 11,
        b: ''
      },
      0.5
    ])
  })

  test('non-object', async () => {
    const v = object({
      a: even,
      b: not(empty)
    })
    expect(await run(v, undefined as any)).toEqual([
      ['Must be an object'],
      { a: 0, b: '' },
      0
    ])
  })

  const movie = object({
    released: atLeast(1888),
    title: both(string, not(empty))
  })

  const book = object({
    author: string,
    ISBN: exactly('ABCD1234')
  })

  test('book ', async () => {
    expect(await run(book, {
      released: 1984,
      title: 'test'
    } as any)).toEqual([
      [
        'author must be a string',
        'ISBN must be exactly ABCD1234'
      ],
      {
        author: '',
        ISBN: 'ABCD1234'
      },
      0
    ])
  })

  test('movie', async () => {
    expect(await run(movie, {
      released: 1983,
      title: 'Return of the Jedi'
    } as any)).toEqual([
      [],
      {
        released: 1983,
        title: 'Return of the Jedi'
      },
      1
    ])
  })

  test('book or movie / movie', async () => {
    const ouvre = either(book, movie)

    expect(await run(ouvre, {
      released: 1983,
      title: 'Return of the Jedi'
    } as any)).toEqual([
      [],
      {
        released: 1983,
        title: 'Return of the Jedi'
      },
      1
    ])
  })

  test('book or movie / book', async () => {
    const ouvre = either(book, movie)

    expect(await run(ouvre, {
      author: 'Stephen King',
      ISBN: 'ABCD1234'
    } as any)).toEqual([
      [],
      {
        author: 'Stephen King',
        ISBN: 'ABCD1234'
      },
      1
    ])
  })

  test('book or movie / validation error', async () => {
    const movie = object({
      released: atLeast(1888),
      title: not(empty)
    })

    const book = object({
      author: string,
      ISBN: exactly('ABCD1234')
    })

    const ouvre = either(book, movie)

    expect(await run(ouvre, {
      author: 123,
      ISBN: 'ABCD1234'
    } as any)).toEqual([
      [
        'author must be a string or released must be a number',
        'author must be a string or title must be a string'
      ],
      {
        author: '',
        ISBN: 'ABCD1234',
      },
      0.5,
    ])
  })

  test('undefined:s in object "compacted"', async () => {
    const schema = object({
      a: optional(atMost(10)),
      b: optional(exactly(2)),
    })
    expect(Object.keys((await run(schema, {}))[1]))
      .toEqual([])
  })

  test('empty object input to validator', async () => {
    const schema = object({})
    expect(await run(schema, { a: 10 })).toEqual([
      [],
      { a: 10 },
      1
    ])
  })

  test('mix object / primitive', async () => {
    const schema = either(number, object({ a: number, b: string }))
    expect(await run(schema, 'test')).toEqual([
      [
        'Must be a number or an object'
      ],
      0,
      0
    ])
  })

  test('mix object / primitive 2', async () => {
    const schema = either(number, object({ a: number, b: string }))
    expect(await run(schema, { a: 1 })).toEqual([
      [
        'Must be a number or an object where b must be a string'
      ],
      {
        a: 1,
        b: ''
      },
      0.5
    ])
  })

  test('mix object / primitive 3', async () => {
    const schema = either(number, object({ a: number, b: either(string, number) }))
    expect(await run(schema, { a: 1 })).toEqual([
      [
        'Must be a number or an object where b must be a string or b must be a number'
      ],
      {
        a: 1,
        b: ''
      },
      0.5
    ])
  })

  test('nested object', async () => {
    const schema = object({
      x: str2num,
      y: either(string, object({
        z: string,
        w: boolean
      }))
    })
    expect(await run(schema, {
      x: '123',
      y: {
        z: '!!',
        w: true
      }
    })).toEqual([
      [],
      { x: 123, y: { z: '!!', w: true } },
      1
    ])
  })
})

describe('length', () => {
  test('ok', async () => {
    const x = object({
      name: length(atLeast(2), atMost(10)),
      ssn: length(exactly(5))
    })

    expect(await run(x, { name: 'Bob', ssn: 'abcde' })).toEqual([
      [],
      { name: 'Bob', ssn: 'abcde' },
      1
    ])
  })

  test('invalid 1', async () => {
    const x = object({
      name: length(atLeast(2), atMost(10)),
      ssn: length(exactly(5))
    })

    expect(await run(x, { name: 'B', ssn: 'abcde' })).toEqual([
      ['name must be of length at least 2'],
      { name: 'B', ssn: 'abcde' },
      0.75
    ])
  })

  test('invalid 2', async () => {
    const x = object({
      name: length(atLeast(2), atMost(10)),
      ssn: length(exactly(5))
    })

    expect(await run(x, { name: 'Bob', ssn: 'abcd' })).toEqual([
      ['ssn must be of length exactly 5'],
      { name: 'Bob', ssn: 'abcd' },
      0.75
    ])
  })
})

describe('self', () => {
  const v = self((obj : { a : number, b : string }) => object({
    a: number,
    b: length(exactly(obj.a))
  }))

  test('ok', async () => {
    expect(await run(v, { a: 5, b: 'abcde' })).toEqual([
      [],
      { a: 5, b: 'abcde' },
      1
    ])
  })

  test('invalid', async () => {
    expect(await run(v, { a: 5, b: 'abcd' })).toEqual([
      ['b must be of length exactly 5'],
      { a: 5, b: 'abcd' },
      0.75
    ])
  })

  test('negated', async () => {
    const w = not(self(x => exactly(x)))
    expect(await run(w, 10)).toEqual([
      ['Must not be exactly 10'],
      10,
      0.5
    ])
  })
})

describe('pipe', () => {
  const dateToDay : Validator<Date, number> = createOneWayParser((d : Date) => {
    const day = d.getDay()
    return [
      day <= 5,
      l => l.custom('a weekday'),
      day
    ]
  })

  test('ok', async () => {
    const schema = pipe(stringToDate, dateToDay)
    expect(await run(schema, '2019-01-01')).toEqual([
      [],
      2,
      1
    ])
  })

  test('invalid first part', async () => {
    const schema = pipe(stringToDate, dateToDay)
    expect(await run(schema, '2019-99-01')).toEqual([
      ['Must be a date'],
      NaN,
      0
    ])
  })

  test('invalid second part', async () => {
    const schema = pipe(stringToDate, dateToDay)
    expect(await run(schema, '2019-01-05')).toEqual([
      ['Must be a weekday'],
      6,
      0
    ])
  })
})

describe('object with negations', () => {
  const movie = object({
    released: numberOf(atLeast(1888)),
    title: stringOf(not(empty))
  })

  const person = object({
    name: stringOf(not(empty)),
    verified: exactly(true)
  })

  const schema =
    either(movie, not(person))

  test('ok / simple', async () => {
    const input = {
      released: 1983,
      title: 'Return of the Jedi'
    }
    expect(await run(schema, input))
      .toEqual([
        [],
        input,
        1
      ])
  })

  test('ok / default', async () => {
    const input = {
      name: '',
      verified: false
    }
    expect(await run(not(person), input))
      .toEqual([
        [],
        {
          name: ''
        },
        1
      ])
  })

  test('3', async () => {
    const input = {
      name: 'something',
      verified: true
    }
    expect(await run(not(person), input))
      .toEqual([
        ['name must be empty or verified must not be exactly true',],
        { name: 'something' },
        0.5
      ])
  })

  test('4', async () => {
    const input = {
      name: 'something',
      verified: true
    }
    expect(await run(schema, input))
      .toEqual([
        [
          'released must be a number or name must be empty or verified must not be exactly true',
          'title must be a string or name must be empty or verified must not be exactly true',
        ],
        {
          name: 'something'
        },
        0.5
      ])
  })
})

describe('async validation', () => {
  function delay(n : number) {
    return new Promise(k => setTimeout(k, n))
  }

  const prime = createSimpleValidator(async function(a : number) {
    await delay(5)
    return [
      (l => l.custom('a prime less than 10')) as Message,
      [2, 3, 5, 7].includes(a),
    ] as [Message, boolean]
  })

  test('simple / valid', async () => {
    expect(await run(prime, 7)).toEqual([[], 7, 1])
  })

  test('simple / invalid', async () => {
    expect(await run(prime, 8)).toEqual([
      ['Must be a prime less than 10'],
      8,
      0.5
    ])
  })

  const countAs = createOneWayParser(async (a : string) => {
    await delay(5)
    const ok = typeof a === 'string'
    return [
      ok,
      l => l.aString(),
      ok ? a.replace(/[^a]/g, '').length : 0,
    ] as [boolean, Message, number]
  })

  test('parser / valid', async () => {
    expect(await run(countAs, 'aabbaac')).toEqual([
      [],
      4,
      1
    ])
  })

  test('parser / invalid', async () => {
    expect(await run(countAs, 123 as any)).toEqual([
      ['Must be a string'],
      0,
      0
    ])
  })

  test('combined / ok 1', async () => {
    expect(await run(either(prime, countAs), "aa")).toEqual([
      [],
      2,
      1
    ])
  })

  test('combined / ok 2', async () => {
    expect(await run(either(prime, countAs), "aa")).toEqual([
      [],
      2,
      1
    ])
  })

  test('either / invalid 1', async () => {
    expect(await run(either(prime, countAs), 9)).toEqual([
      ['Must be a prime less than 10 or a string'],
      9,
      0.5
    ])
  })

  test('pipe / ok', async () => {
    expect(await run(pipe(countAs, prime), 'aaa')).toEqual([
      [],
      3,
      1
    ])
  })

  test('pipe / invalid 1', async () => {
    expect(await run(pipe(countAs, prime), 3 as any)).toEqual([
      ['Must be a string'],
      0,
      0
    ])
  })

  test('pipe / invalid 2', async () => {
    expect(await run(pipe(countAs, prime), 'aaaa')).toEqual([
      ['Must be a prime less than 10'],
      4,
      0.5
    ])
  })

  test('both / ok', async () => {
    expect(await run(both(prime, even), 2)).toEqual([
      [],
      2,
      1
    ])
  })

  test('both / invalid', async () => {
    expect(await run(both(prime, even), 3)).toEqual([
      ['Must be even'],
      3,
      0.75
    ])
  })

  test('object / ok', async () => {
    expect(await run(
      object({ i: prime, c: countAs }),
      { i: 7, c: '8aabbcca' }
    )).toEqual([
      [],
      { i: 7, c: 3 },
      1
    ])
  })

  test('object / invalid', async () => {
    expect(await run(
      object({ i: prime, c: pipe(countAs, even) }),
      { i: 6, c: '8aabbcca' }
    )).toEqual([
      [
        'i must be a prime less than 10',
        'c must be even'
      ],
      { i: 6, c: 3 },
      0.5
    ])
  })
})

describe('errors per key', () => {

  const user = object({
    name: optional(string),
    password: stringOf(length(atLeast(8))),
    age: numberOf(atLeast(18))
  })

  const signupData = object({
    user,
    terms: object({
      accepted: exactly(true),
      receiveNewsletter: either(exactly(true), exactly(false))
    })
  })

  test('object', async () => {
    const l = describePaths(enUS, [
      ['*.password', 'Password'],
      ['*.age', 'Age']
    ])

    expect(await errorsPerKey(user, {}, l)).toEqual({
      password: 'Password must be a string',
      age: 'Age must be a number'
    })
  })

  test('nested object', async () => {
    const l = describePaths(enUS, [
      ['user.password', 'Password'],
      ['*.age', 'Age'],
      ['*.accepted', 'Accepted'],
      ['terms.receiveNewsletter', 'Receive Newsletter']
    ])

    expect(await errorsPerKey(signupData, { user: {}, terms: {} }, l)).toEqual({
      user: {
        password: 'Password must be a string',
        age: 'Age must be a number'
      },
      terms: {
        accepted: 'Accepted must be exactly true',
        receiveNewsletter: 'Receive Newsletter must be exactly true or Receive Newsletter must be exactly false'
      }
    })
  })

  test('non-object', async () => {
    expect(await errorsPerKey(number, null, enUS)).toEqual({})
  })

  const v = either(number, object({ n: number }))

  test('mixed-object', async () => {
    expect(await errorsPerKey(v, null, enUS)).toEqual({
      n: 'Must be a number or an object where n must be a number'
    })
  })
})

describe('implications', () => {
  test('invalid / 1', async () => {
    const v = when(even, moreThan(10))
    expect(await run(v, 8)).toEqual([
      ['Must be more than 10 when even'],
      8,
      0.5
    ])
  })

  test('invalid / disj', async () => {
    const v = when(either(even, atLeast(5)), moreThan(10))
    expect(await run(v, 9)).toEqual([
      ['Must be more than 10 when at least 5'],
      9,
      0.5
    ])
  })

  test('invalid / conj', async () => {
    const v = when(both(even, atLeast(5)), moreThan(10))
    expect(await run(v, 6)).toEqual([
      ['Must be more than 10 when even and at least 5'],
      6,
      0.5
    ])
  })
})


