import { displayOneOf } from "./utils";

export interface Language<I extends Rep> {
  aString() : I
  aNumber() : I
  aBoolean() : I
  anObject() : I
  oneOf(i : unknown[]) : I
  iterable() : I
  iterableOf(of : I) : I
  anArray() : I
  anArrayOf(of : I) : I
  aList() : I
  date() : I
  empty() : I
  alphaNum() : I
  atMost(ub : number) : I
  atLeast(lb : number) : I
  moreThan(lb : number) : I
  lessThan(ub : number) : I
  even() : I
  exactly(v : any) : I
  not(v : I) : I
  atKey(key : string, v : I) : I
  lengthOf(v : I) : I
  custom(v : string) : I
  eitherOr(msg : I[]) : I
  self() : I
  describe(i : I[]) : string
  finalize(i : I[]) : string
  when(cond : I[][], cons : I[]) : I
  describePath(p : Path) : string
}

export type Message =
  <I extends Rep>(lang : Language<I>) => I

type Path = string[]

export type Rep = {
  neg : boolean,
  pth : Path,
  msg : string
}

function msg(msg : string) : Rep {
  return { msg, neg: false, pth: [] }
}

function capFirst(s : string) {
  return s.charAt(0).toUpperCase() + s.slice(1);
}

export const enUS : Language<Rep> = {
  aString() : Rep {
    return msg('a string')
  },

  aNumber() : Rep {
    return msg('a number')
  },

  aBoolean() : Rep {
    return msg('true or false')
  },

  anObject() : Rep {
    return msg('an object')
  },

  oneOf(choices : unknown[]) {
    const strs = choices.map(x => `${x}`)
    return msg(`one of ${displayOneOf(strs.map(c => this.describePath([c])), ' or ')}`)
  },

  iterable() : Rep {
    return msg('iterable')
  },

  iterableOf({ neg, pth, msg } : Rep) {
    return { neg, pth, msg: `an iterable where each element is ${msg}` }
  },

  anArray() : Rep {
    return msg('an array')
  },

  anArrayOf({ neg, pth, msg } : Rep) {
    return { neg, pth, msg: `an array where each element is ${msg}` }
  },

  aList() : Rep {
    return msg('a list')
  },

  date() : Rep {
    return msg('a date')
  },

  empty() : Rep {
    return msg('empty')
  },

  alphaNum() : Rep {
    return msg('alphanumeric')
  },

  atMost(ub : number) {
    return msg(`at most ${ub}`)
  },

  atLeast(lb : number) {
    return msg(`at least ${lb}`)
  },

  moreThan(lb : number) {
    return msg(`more than ${lb}`)
  },

  lessThan(ub : number) {
    return msg(`less than ${ub}`)
  },

  even() : Rep {
    return msg('even')
  },

  exactly(v : any) : Rep {
    return msg(`exactly ${v}`)
  },

  not({ neg, pth, msg } : Rep) : Rep {
    return { neg: !neg, pth, msg }
  },

  atKey(key : string, { neg, pth, msg } : Rep) : Rep {
    return { neg, pth: [key, ...pth], msg }
  },

  lengthOf({ neg, pth, msg } : Rep) : Rep {
    return { neg, pth, msg: `of length ${msg}` }
  },

  custom(m : string) : Rep {
    return msg(m)
  },

  self() : Rep {
    return msg('of this schema type')
  },

  eitherOr(msg : Rep[]) : Rep {
    return {
      neg: false,
      pth: [],
      msg: this.describe(msg)
    }
  },

  describe(rs : Rep[]) : string {
    const es = rs.filter(x => x.pth.length === 0)
    const ps = rs.filter(x => x.pth.length > 0)
    const poss = es.filter(e => !e.neg)
    const negs = es.filter(e => e.neg)
    let clauses = []
    if (poss.length) {
      clauses.push(`${poss.map(x => x.msg).join(' or ')}`)
    }
    if (negs.length) {
      clauses.push(`not ${negs.map(x => x.msg).join(' or ')}`)
    }
    if (ps.length) {
      clauses.push(`an object where ${this.finalize(ps)}`)
    }
    return clauses.join(' or ')
  },

  finalize(rs : Rep[]) : string {
    const es = rs.filter(x => x.pth.length === 0)
    const ps = rs.filter(x => x.pth.length > 0)
    if (es.length > 0) {
      const poss = es.filter(e => !e.neg)
      const negs = es.filter(e => e.neg)
      let clauses = []
      if (poss.length) {
        clauses.push(`must be ${poss.map(x => x.msg).join(' or ')}`)
      }
      if (negs.length) {
        clauses.push(`must not be ${negs.map(x => x.msg).join(' or ')}`)
      }
      if (ps.length) {
        clauses.push(`an object where ${this.finalize(ps)}`)
      }
      return capFirst(clauses.join(' or '))
    }
    return ps
      .map(({ neg, pth, msg }) => {
        if (neg) {
          return `${this.describePath(pth)} must not be ${msg}`
        }
        return `${this.describePath(pth)} must be ${msg}`
      })
      .join(' or ')
  },

  when(cond : Rep[][], conseq : Rep[]) : Rep {
    return {
      neg: false,
      pth: [],
      msg: `${this.describe(conseq)} when ${cond.map(d => this.describe(d)).join(' and ')}`
    }
  },

  describePath(path) {
    return path.join('.')
  }
}

export function describePaths<I extends Rep>(
  l : Language<I>,
  paths : [string, string][]
) : Language<I> {
  const exprs : [RegExp, string][] =
    paths.map(([expr, translation]) => {
      return [
        new RegExp(
          '^'
          + expr.replace(/\*/g, '\\w*').replace(/\./g, '\\.?')
          + '$'
        ),
        translation
      ]
    })
  return {
    ...l,
    describePath(path : Path) : string {
      const p = path.join('.')
      const c = exprs.find(x => x[0].test(p))
      if (c !== undefined) return c[1]
      return p
    }
  }
}

