import {
  Err,
  Validator,
  both as bothV,
  either as eitherV,
  not as notV,
  id as idValidator,
  invalidDate,
  object as objectV,
  string as stringV,
  UndefinedOptional,
  stringOf as _stringOf,
  numberOf as _numberOf,
  createValidator,
  mapValidator,
  contramapValidator,
  createSimpleValidator,
  checkType,
  createOneWayParser,
  vErr
} from './validators'
import { _true } from './logic'
import { Message } from './message'

export type Schema<A, B> = {
  parse : Validator<A, B>,
  unparse : Validator<B, A>,
}

export function both<A, B, C, D>(
  v : Schema<A, C>,
  w : Schema<B, D>,
) : Schema<A & B, C & D> {
  return {
    parse: bothV(v.parse, w.parse),
    unparse: bothV(v.unparse, w.unparse)
  }
}

export function either<A, B, C, D>(
  v : Schema<A, C>,
  w : Schema<B, D>
) : Schema<A | B, C | D> {
  return {
    parse: eitherV(v.parse, w.parse),
    unparse: eitherV(v.unparse, w.unparse)
  }
}

export function notSchema<A, B>(v : Schema<A, B>) : Schema<A, B> {
  return {
    parse: notV(v.parse),
    unparse: notV(v.unparse)
  }
}

export function mapObject<AS, B>(
  a : AS,
  f : (a : (AS[keyof AS])) => B
) : { [key in keyof AS]: B } {
  let result : { [key in keyof AS]: B } = {} as any
  const keys = Object.keys(a)
  for (const i in keys) {
    const key = keys[i]
    // @ts-ignore
    result[key] = f(a[key])
  }
  return result
}

export function object<AS, BS>(
  vs :
    { [Key in keyof AS]: Schema<AS[Key], any> } &
    { [Key in keyof BS]: Schema<any, BS[Key]> }
) : Schema<UndefinedOptional<AS>, UndefinedOptional<BS>> {
  return {
    parse: objectV(mapObject(vs, (x : any) => x.parse)) as any,
    unparse: objectV(mapObject(vs, (x : any) => x.unparse)) as any
  }
}

export const irreversible = (func : string) => () => {
  throw new Error(
    `Schema is not reversible because ${func} was used without its second argument.`
  )
}

export function createSchema<A, B>(
  parse : (a : A, inv : boolean) => [boolean, Err, B, boolean?],
  unparse : (b : B, inv : boolean) => [boolean, Err, A, boolean?] = irreversible('createSchema'),
) : Schema<A, B> {
  return {
    parse: createValidator(parse),
    unparse: createValidator(unparse)
  }
}

export function createSimpleSchema<A>(
  parse : (a : A) => [Message, boolean] | Promise<[Message, boolean]>,
  unparse : (a : A) => [Message, boolean] | Promise<[Message, boolean]> = irreversible('createSimpleSchema'),
) : Schema<A, A> {
  return {
    parse: createSimpleValidator(parse),
    unparse: createSimpleValidator(unparse)
  }
}

export function map<A, B, C>(
  s : Schema<A, B>,
  f : (b : B) => C,
  g : (c : C) => B = irreversible('map')
) : Schema<A, C> {
  return {
    parse: mapValidator(s.parse, f),
    unparse: contramapValidator(s.unparse, g)
  }
}

export function contramap<A, B, C>(
  s : Schema<A, B>,
  f : (c : C) => A,
  g : (b : A) => C = irreversible('contramap')
) : Schema<C, B> {
  return {
    parse: contramapValidator(s.parse, f),
    unparse: mapValidator(s.unparse, g)
  }
}

export function reverse<A, B>(schema : Schema<A, B>) : Schema<B, A> {
  return { parse: schema.unparse, unparse: schema.parse }
}

export const id : <A>() => Schema<A, A> = () => ({
  parse: idValidator,
  unparse: idValidator
})

export const string : Schema<string, string> = {
  parse: stringV,
  unparse: stringV
}

export const stringOf = checkInputType(
  (a : unknown) => [
    typeof a === 'string',
    typeof a === 'string' ? a : '',
    l => l.aString()
  ]
)

export const numberOf = checkInputType(
  (a : unknown) => [
    typeof a === 'number',
    typeof a === 'number' ? a : 0,
    l => l.aNumber()
  ]
)

export const even = numberOf(createSimpleSchema((a : number) => [
  l => l.even(),
  a % 2 === 0
]))

export function checkInputType<A>(
  checkL : (a : unknown) => [boolean, A, Message]
) {
  return function <B = A>(
    v : Schema<A, B> = id as unknown as Schema<A, B>
  ) : Schema<A, B> {
    return {
      parse: checkType(checkL)(v.parse),
      unparse: v.unparse
    }
  }
}

export function oneOf<A>(
  ...choices : A[]
) {
  function oneof(a : A) {
    const ok = choices.includes(a)
    return [
      ok,
      vErr(l => l.oneOf(choices), ok),
      ok ? a : choices.length ? choices[0] : a,
    ] as [boolean, Err, A]
  }
  return createSchema(oneof, oneof)
}

export const empty : Schema<string, string> = stringOf(
  createSimpleSchema(s => [
    l => l.empty(),
    s !== ''
  ]))

export const exactly = <A>(
  target : A,
  equals : (x : A, y : A) => boolean = (x, y) => x === y
) : Schema<A, A> => {
  const fn = (a : unknown) => {
    const typeOk = typeof a === typeof target
    const ok = typeOk && equals(a as A, target)
    return [
      typeOk,
      vErr(l => l.exactly(target), ok),
      ok ? (a as A) : target,
      a === target
    ] as [boolean, Err, A, boolean]
  }
  return createSchema(fn, fn)
}

export function atLeast(lb : number) : Schema<any, number> {
  return numberOf(createSimpleSchema(a => [
    l => l.atLeast(lb),
    a >= lb
  ]))
}

export function createParser<A, B>(
  parse : (a : A) => [boolean, Message, B],
  unparse : (a : B) => [boolean, Message, A] = irreversible('createParser')
) {
  return {
    parse: createOneWayParser(parse),
    unparse: createOneWayParser(unparse)
  }
}

export const str2num = {
  parse: _stringOf(createOneWayParser(
    (s : string) => {
      const result = Number(s)
      const ok = s !== '' && !isNaN(result)
      return [
        ok,
        l => l.aNumber(),
        ok ? result : 0
      ]
    }
  )),
  unparse: _numberOf(createOneWayParser((n : number) => [
    typeof n === 'number',
    l => l.aNumber(),
    typeof n === 'number' ? `${n}` : ''
  ]))
}

// export const alphaNumeric : Validator<string, string> = function(s : string) {
//   const isstr = typeof s === 'string'
//   const ok = isstr && s.match(/^[a-zA-Z0-9]*$/) != null
//   return [
//     [[{ v: l => l.alphaNum(), ok }]],
//     typeof s === 'string' ? s : '',
//     isstr ? 1 : 0
//   ]
// }

export const str2date = stringOf(createParser(
  (s : string) => {
    const date = new Date(s)
    const ok = !isNaN(date.getTime())
    return [
      ok,
      l => l.date(),
      ok ? date : invalidDate
    ]
  },
  (date : Date) => {
    const str =
      date instanceof Date ? date.toISOString().split('T')[0] : ''
    const ok =
      date instanceof Date && !isNaN(date.getTime())
    return [
      ok,
      l => l.date(),
      str
    ]
  }
))
