import { Message, enUS, Rep, Language } from './message'
import { CNF, and, or, _true } from './logic'
import Action from './action'

export type Err = {
  k : 'v' | 'p',
  m : Message,
  ok : boolean
}

export function vErr(m : Message, ok : boolean) : Err {
  return { k: 'v', m, ok }
}

export function pErr(m : Message, ok : boolean) : Err {
  return { k: 'p', m, ok }
}

function evalDisjunction<I extends Rep>(
  l : Language<I>
) {
  return function evalDisjInner(
    msg : Err[],
  ) : I[] {
    return msg.map(m => m.m(l))
  }
}

export type Validator<A, B> =
  (a : A, inverted : boolean) =>
    Action<[CNF<Err>, B, number]>

export function merge<A, B>(a : A, b : B) : A & B {
  if (
    typeof a === 'object' &&
    typeof b === 'object'
  ) {
    return { ...a, ...b }
  }
  return a as A & B
}

export function both<A, B, C, D>(
  v : Validator<A, C>,
  w : Validator<B, D>
) : Validator<A & B, C & D> {
  return function(a : A & B, inv : boolean) {
    if (inv) {
      return either(not(v), not(w))(a, false) as any
    }
    return v(a, false).cont(([cnfC, c, okC]) => {
      return w(a, false).map(([cnfD, d, okD]) => {
        return [
          and(cnfC, cnfD),
          okC >= okD ? merge(c, d) : merge(d, c),
          (okC + okD) / 2
        ]
      })
    })
  }
}

export function either<A, B, C, D>(
  v : Validator<A, C>,
  w : Validator<B, D>
) : Validator<A | B, C | D> {
  return function(a : A | B, inv : boolean) {
    if (inv) {
      return both(not(v), not(w))(a as any, false) as any
    }
    return v(a as any, false).cont(([cnfC, c, okC]) => {
      return w(a as any, false).map(([cnfD, d, okD]) => {
        return [
          or(cnfC, cnfD),
          okC >= okD ? c : d,
          Math.max(okC, okD)
        ]
      })
    })
  }
}

export function not<A, B>(
  v : Validator<A, B>
) : Validator<A, B> {
  return function(a : A, inv : boolean) {
    return v(a, !inv).map(([cnf, b, score]) => [
      cnf,
      b,
      score
    ])
  }
}

export function when<A, B>(
  cond : Validator<A, B>,
  consequent : Validator<A, any>,
  alternative?: Validator<A, any>
) : Validator<A, B> {
  return function(a : A, inv : boolean) {
    return cond(a, inv).cont(([d, b, ok]) => {
      if (ok === 1) {
        const reason : CNF<Err> =
          d.map(disj => disj.filter(x => x.ok))
            .filter(x => x.length)
        return consequent(a, inv).map(([e, , ok]) => {
          return [
            e.map(disj => [{
              k: 'v',
              m: l => l.when(
                reason.map(evalDisjunction(l)),
                evalDisjunction(l)(disj)
              ),
              ok: false
            }]),
            b,
            ok
          ]
        })
      }
      return alternative
        ? alternative(a, inv).map(([e, , ok]) => {
          return [e, b, ok]
        })
        : new Action([[], b, ok])
    })
  }
}

export function id<A>(a : A) {
  return new Action(
    [[], a, 1] as [CNF<Err>, A, number]
  )
}

export const stringOf = checkType(
  (a : unknown) => [
    typeof a === 'string',
    typeof a === 'string' ? a : '',
    l => l.aString()
  ]
)

export const numberOf = checkType(
  (a : unknown) => [
    typeof a === 'number',
    typeof a === 'number' ? a : 0,
    l => l.aNumber()
  ]
)

export const booleanOf = checkType(
  (a : unknown) => [
    typeof a === 'boolean',
    typeof a === 'boolean' ? a : 0,
    l => l.aBoolean()
  ]
)

export const string : Validator<unknown, string> = stringOf()

export const number : Validator<unknown, number> = numberOf()

export const boolean : Validator<unknown, number> = booleanOf()

export function handleInv(inv : boolean, m : Message) : Message {
  return l => inv ? l.not(m(l)) : m(l);
}

export const exactly = <A>(
  target : A,
  equals : (x : A, y : A) => boolean = (x, y) => x === y
) : Validator<unknown, A> => {
  return createValidator(a => {
    const ok = equals(a as A, target)
    return [
      typeof a === typeof target,
      vErr(l => l.exactly(target), ok),
      ok ? a as A : target,
      a === target
    ]
  })
}

export function mapValidator<A, B, C>(
  v : Validator<A, B>,
  f : (b : B) => C
) : Validator<A, C> {
  return function(a : A, inv : boolean) {
    return v(a, inv).map(([e, b, ok]) => {
      return [e, f(b), ok]
    })
  }
}

export function contramapValidator<A, B, C>(
  v : Validator<A, B>,
  f : (c : C) => A
) : Validator<C, B> {
  return function(a : C, inv : boolean) {
    return v(f(a), inv)
  }
}

export function createValidator<A, B>(
  check : (a : A, inv : boolean) =>
    [boolean, Err, B, boolean?] | Promise<[boolean, Err, B, boolean?]>,
) : Validator<A, B> {
  return function(a : A, inv : boolean) {
    return new Action(check(a, inv)).cont(([ok, msg, b, score]) => {
      const good = (ok && score !== false) !== inv
      const points = ok ? ((score !== inv) ? 1 : 0.5) : 0
      return new Action([
        [[{
          k: msg.k,
          m: handleInv(inv, msg.m),
          ok: good
        }]],
        b,
        points
      ])
    })
  }
}

export function createSimpleValidator<A>(
  check : (a : A) => [Message, boolean] | Promise<[Message, boolean]>
) : Validator<A, A> {
  return createValidator((a : A) => {
    return new Action(check(a)).map(([err, ok]) => {
      return [true, vErr(err, ok), a, ok] as [boolean, Err, A, boolean?]
    }).val
  })
}

export function createOneWayParser<A, B>(
  parse : (a : A) => [boolean, Message, B] | Promise<[boolean, Message, B]>,
) : Validator<A, B> {
  return createValidator((a : A) => {
    return new Action(parse(a)).map(([ok, err, b]) => {
      return [ok, pErr(err, ok), b] as [boolean, Err, B, boolean?]
    }).val
  })
}

export function checkType<A>(
  check : (a : unknown) => [boolean, A, Message] | Promise<[boolean, A, Message]>
) {
  return function <B>(v : Validator<A, B> = id as any) : Validator<unknown, B> {
    return function(u : unknown, inv : boolean) {
      return new Action(check(u)).cont(([ok, a, msg]) => {
        return v(a, inv).map(
          x => [
            ok ? x[0] : [[pErr(msg, ok)]],
            x[1],
            ok ? x[2] : 0
          ])
      })
    }
  }
}

export const even = numberOf(createSimpleValidator((a : number) => [
  l => l.even(),
  a % 2 === 0
]))

export function atMost(ub : number) : Validator<unknown, number> {
  return numberOf(createSimpleValidator(a => [
    l => l.atMost(ub),
    a <= ub,
  ]))
}

export function atLeast(lb : number) : Validator<any, number> {
  return numberOf(createSimpleValidator(a => [
    l => l.atLeast(lb),
    a >= lb
  ]))
}

export function moreThan(lb : number) : Validator<any, number> {
  return numberOf(createSimpleValidator(a => [
    l => l.moreThan(lb),
    a > lb
  ]))
}

export function lessThan(lb : number) : Validator<any, number> {
  return numberOf(createSimpleValidator(a => [
    l => l.lessThan(lb),
    a < lb
  ]))
}

export const empty = stringOf(createSimpleValidator(s => [
  l => l.empty(),
  s === ''
]))

export const str2num = stringOf(createOneWayParser((s : string) => {
  const result = Number(s)
  const ok = s !== '' && !isNaN(result)
  return [
    ok,
    l => l.aNumber(),
    ok ? result : 0
  ]
}))

export const invalidDate = new Date('')

export const stringToDate = stringOf(createOneWayParser((s : string) => {
  const date = new Date(s)
  const ok = !isNaN(date.getTime())
  return [
    ok,
    l => l.date(),
    ok ? date : invalidDate
  ]
}))

export const checkArray = checkType((a : unknown) => [
  a instanceof Array,
  a instanceof Array ? a : [],
  l => l.anArray()
])

export const arrayOf = <A, B>(v : Validator<A, B>) =>
  checkArray(iterableOf(v, m => l => l.anArrayOf(m(l))))

export function isIterable(a : any) : a is Iterable<any> {
  return !!a && typeof a[Symbol.iterator] === 'function'
}

export const checkIterable = checkType(as => [
  isIterable(as),
  isIterable(as) ? as : [],
  l => l.iterable()
])

export function iterableOf<A, B>(
  v : Validator<A, B>,
  msg : (m : Message) => Message = m => l => l.iterableOf(m(l))
) : Validator<Iterable<A>, B[]> {
  return checkIterable(function(as : Iterable<A>, inv : boolean) {
    const arr = [...as]
    return Action.all(arr.map(a => v(a, inv))).map(results => {
      const bs = results.map(x => x[1])
      const es = results.find(x => x[2] !== 1)
      return [
        [[
          vErr(
            msg(l => {
              const [cnf,] = es!
              return l.eitherOr(evalDisjunction(l)(cnf[0]))
            }),
            !(es && es[0].length)
          )
        ]],
        bs,
        arr.length === 0
          ? 1
          : results.reduce((x, y) => x + y[2], 0) / results.length
      ]
    })
  })
}

export function self<A, B>(fn : (a : A) => Validator<A, B>) : Validator<A, B> {
  return function(a : A, inv : boolean) {
    return fn(a)(a, inv)
  }
}

export function setMessage<A, B>(
  v : Validator<A, B>,
  msg : Message
) : Validator<A, B> {
  return function(a : A, inv : boolean) {
    return v(a, inv).map(([, b, ok]) => {
      return [[[vErr(msg, ok === 1)]], b, ok]
    })
  }
}

export function fix<A, B = A>(fn : (v : Validator<A, B>) => Validator<A, B>) : Validator<A, B> {
  let x : Validator<A, B> = () => {
    throw new Error('Strict usage of `fix`')
  }
  const check : Validator<A, B> = setMessage(
    (a : A, inv : boolean) => {
      return x(a, inv)
    },
    l => l.self()
  )
  return x = fn(check)
}

export function optional<A, B>(
  v : Validator<A, B>
) : Validator<A | null | void, B | null | void> {
  return function(a : A | null | void, inv : boolean) {
    if (a === null || a === undefined) {
      return new Action([
        _true,
        a as null | void,
        1
      ])
    }
    return v(a, inv)
  }
}

export type UndefinedOptional<A> =
  { [P in keyof A]?: A[P] } &
  { [L in { [K in keyof A]: undefined extends A[K] ? never : K }[keyof A]]: A[L] }

export const checkObject = checkType(o => [
  typeof o === 'object',
  (typeof o === 'object' ? o : {}) as object,
  l => l.anObject()
])

export function atKey<A, B, K extends string>(
  key : K,
  v : Validator<A, B>
) : Validator<{ [k in K]: A }, { [k in K]: B }> {
  return function(a : { [k in K]: A }, inv : boolean) {
    return v((a || {})[key], inv).map(([e, b, ok]) => {
      const e_ = e.map(c => c.map(
        d => ({
          k: d.k,
          m: l => l.atKey(key as string, d.m(l)),
          ok: d.ok
        }) as Err
      ))
      return [
        e_,
        (b === undefined ? {} : { [key]: b }) as { [k in K]: B },
        ok
      ]
    })
  }
}

export function object<BS>(
  vs : { [Key in keyof BS]: Validator<any, BS[Key]> }
) : Validator<unknown, UndefinedOptional<BS>> {
  const keys = Object.keys(vs) as (keyof BS)[]
  return checkObject(
    keys.length === 0 ? id :
      keys
        .map(key => atKey(key as string, vs[key]))
        .reduce(both) as any
  )
}

export function length<A>(...vs : Validator<number, any>[]) : Validator<Iterable<A>, Iterable<A>> {
  return function(a : Iterable<A>, inv : boolean) {
    return Action.all(vs.map(v => v([...a].length, inv))).map(results => {
      const es = results
        .map(r => r[0]).reduce(and, [])
        .map(d => d.map(e => ({
          k: e.k,
          m: l => l.lengthOf(e.m(l)),
          ok: e.ok
        }) as Err))
      const ok = Math.min(...results.map(r => r[2]))
      return [
        and(es, [[pErr(l => l.iterable(), isIterable(a))]]),
        a,
        ok
      ]
    })
  }
}

export async function run<A, B>(v : Validator<A, B>, a : A) : Promise<[string[], B, number]> {
  const [cnf, b, p] = await v(a, false).val
  const simplified = cnf
    .filter(x => !x.some(y => y.ok))
  const msg = simplified.map(disj => {
    return enUS.finalize(disj.map(x => x.m(enUS)))
  })
  return [msg, b, p]
}

type Messages = string | {
  [k in string]: Messages
}

function rec(msg : Rep[][], l : Language<any>) : Messages {
  const result : Messages = {}
  msg.forEach(clause => {
    clause.forEach(disj => {
      let obj : Messages = result
      const p = disj.pth
      for (let i = 0; i < p.length - 1; i++) {
        // @ts-ignore
        obj[p[i]] = obj[p[i]] || {}
        // @ts-ignore
        obj = obj[p[i]]
      }
      if (p.length > 0) {
        // @ts-ignore
        obj[p[p.length - 1]] = l.finalize(clause)
      }
    })
  })
  return result
}

export async function errorsPerKey<A, B>(
  v : Validator<A, B>,
  a : A,
  l : Language<any>
) : Promise<Messages> {
  const [cnf] = await v(a, false).val
  const msg = cnf.map(disj => {
    return disj.map(x => x.m(enUS))
  })
  return rec(msg, l)
}

export function cnfOK(e : CNF<Err>) : boolean {
  return e.every(disj => disj.some(x => x.ok))
}

export function pipe<A, B, C>(
  v : Validator<A, B>,
  w : Validator<B, C>
) : Validator<A, C> {
  return function(a : A, inv : boolean) {
    return v(a, inv).cont(([e, b, ok1]) => {
      return w(b, inv).map(([f, c, ok2]) => {
        if (!cnfOK(e)) {
          return [e, c, ok1]
        }
        return [f, c, ok2]
      })
    })
  }
}
